use std::{
    sync::mpsc::{self, Receiver, Sender, SyncSender},
    thread,
    time::{Duration, Instant},
};

use anyhow::Context;

use crate::{
    aprs::CommandHandler,
    config::Config,
    img::ImgManager,
    packet::{FecPacket, Packet, RawPacket},
    radio::McuRadio,
    ssdv::ssdv_encode,
    video::{start_video, VideoPacker},
};

const IMAGE_PACKET_QUEUE_LENGTH: usize = 8192;
const VIDEO_PACKET_QUEUE_LENGTH: usize = 1024;
const TEMP_REFRESH_INTERVAL: Duration = Duration::from_secs(5);

pub struct Controller {
    config: Config,
}

impl Controller {
    pub fn new(config: Config) -> Self {
        Self { config }
    }

    pub fn run_forever(self) {
        let (img_tx, img_rx) = mpsc::sync_channel(IMAGE_PACKET_QUEUE_LENGTH);
        let (vid_tx, vid_rx) = crossbeam_channel::bounded(VIDEO_PACKET_QUEUE_LENGTH);
        let (telem_tx, telem_rx) = mpsc::channel();
        let (cmd_tx, cmd_rx) = mpsc::channel();

        {
            let callsign = self.config.callsign.clone();
            thread::spawn(|| Self::tx_thread(callsign, img_rx, vid_rx, telem_rx));
        }

        {
            let config = self.config.clone();
            thread::spawn(|| Self::aprs_thread(config, cmd_tx, telem_tx));
        }

        {
            start_video(vid_tx);
        }

        let mut manager = ImgManager::new(self.config.paths.clone(), cmd_rx);

        loop {
            while let Some((idx, bytes)) = manager.next(self.config.max_img_dimension) {
                if let Err(e) = self.process_image(&bytes, idx, &img_tx) {
                    eprintln!("Error processing image: {e}");
                }
            }

            // after we don't find anything, sleep for a bit while
            // we wait for new images
            thread::sleep(Duration::from_secs(1));
        }
    }

    fn process_image(&self, img: &[u8], idx: u8, tx: &SyncSender<FecPacket>) -> anyhow::Result<()> {
        for p in ssdv_encode(&self.config.callsign, img, idx)? {
            tx.send(p.into()).context("TX thread died")?;
        }

        Ok(())
    }

    // manages incoming APRS packets
    // used to request HD images, as well as
    // to initiate burst/cutdown
    fn aprs_thread(config: Config, tx: Sender<u8>, telem_tx: Sender<Packet>) {
        if let Some(ctrl) = config.control {
            let mut handler = CommandHandler::new(
                config.callsign,
                ctrl.secret,
                ctrl.burst_command,
                ctrl.cutdown_command,
                tx,
                telem_tx,
            );

            handler.process_forever();
        }
    }

    // manages our transceiver
    fn tx_thread(
        callsign: String,
        image_rx: Receiver<FecPacket>,
        vid_rx: crossbeam_channel::Receiver<Vec<u8>>,
        telem_rx: Receiver<Packet>,
    ) {
        let mut radio = loop {
            let r = McuRadio::new();

            match r {
                Ok(r) => break r,
                Err(e) => {
                    eprintln!("Error initializing radio: {e}");
                }
            }

            thread::sleep(Duration::from_millis(1000));
        };

        let mut text_msg_id = 0;
        let mut video_packer = VideoPacker::new();
        let mut last_got_temp = Instant::now();
        loop {
            if let Err(e) = Self::tx_thread_single_iter(
                &callsign,
                &image_rx,
                &vid_rx,
                &telem_rx,
                &mut text_msg_id,
                &mut last_got_temp,
                &mut video_packer,
                &mut radio,
            ) {
                eprintln!("Radio error: {}", e);
                radio.reset();
            }
        }
    }

    // TODO this should probably be a struct
    #[allow(clippy::too_many_arguments)]
    fn tx_thread_single_iter(
        callsign: &str,
        image_rx: &Receiver<FecPacket>,
        vid_rx: &crossbeam_channel::Receiver<Vec<u8>>,
        telem_rx: &Receiver<Packet>,
        text_msg_id: &mut u16,
        last_got_temp: &mut Instant,
        video_packer: &mut VideoPacker,
        radio: &mut McuRadio,
    ) -> anyhow::Result<()> {
        while let Ok(tm) = telem_rx.try_recv() {
            let tm = tm.into_raw(text_msg_id).into();

            radio.send_packet(&tm)?;
        }

        while let Ok(vid) = vid_rx.try_recv() {
            video_packer.pack(&vid, |pkt| radio.send_packet(&pkt.into()))?;
        }

        if let Ok(img) = image_rx.try_recv() {
            radio.send_packet(&img)?;
        } else {
            // send garbage. We want our radio to be active at all times, for AGC purposes on the downlink side
            const GARBAGE: [u8; 256] = [
                102, 54, 91, 67, 89, 223, 198, 83, 22, 16, 48, 184, 117, 97, 153, 246, 169, 167,
                89, 85, 49, 67, 128, 123, 83, 210, 58, 104, 248, 102, 219, 195, 121, 57, 101, 172,
                57, 223, 190, 106, 90, 36, 39, 156, 99, 92, 87, 10, 56, 29, 137, 71, 144, 89, 82,
                182, 127, 72, 93, 249, 214, 6, 155, 164, 177, 22, 84, 111, 52, 60, 68, 235, 30, 13,
                174, 101, 49, 43, 95, 61, 214, 89, 110, 24, 77, 208, 103, 209, 87, 12, 218, 147,
                224, 85, 178, 49, 28, 233, 65, 132, 61, 238, 70, 164, 177, 90, 158, 99, 180, 77,
                251, 17, 227, 43, 109, 33, 120, 15, 89, 172, 69, 213, 25, 166, 59, 254, 220, 31,
                21, 247, 246, 12, 204, 223, 134, 136, 100, 92, 20, 182, 204, 79, 239, 120, 8, 40,
                138, 222, 239, 85, 15, 196, 169, 36, 38, 193, 207, 165, 7, 4, 33, 4, 120, 250, 114,
                240, 128, 3, 22, 62, 254, 139, 13, 56, 153, 15, 63, 96, 62, 44, 128, 241, 25, 22,
                125, 127, 0, 137, 165, 145, 156, 39, 90, 94, 145, 86, 156, 17, 187, 217, 249, 193,
                112, 160, 238, 216, 183, 46, 27, 74, 38, 127, 233, 188, 184, 35, 194, 249, 90, 195,
                33, 21, 67, 56, 75, 243, 140, 6, 187, 93, 49, 224, 20, 34, 204, 204, 141, 132, 252,
                101, 3, 149, 107, 173, 139, 125, 41, 133, 251, 42, 171, 130, 254, 145, 34, 56,
            ];

            radio.send_packet(&RawPacket(GARBAGE).into())?;
        }

        if Instant::now() - *last_got_temp > TEMP_REFRESH_INTERVAL {
            *last_got_temp = Instant::now();

            let temp = radio.get_temp()?;

            let packet = Packet::new_text_message(
                callsign,
                &format!(
                    "Temp: {}, Vbuf: {} / {}",
                    temp,
                    vid_rx.len(),
                    VIDEO_PACKET_QUEUE_LENGTH
                ),
            );

            radio.send_packet(&packet.into_raw(text_msg_id).into())?;
        }

        Ok(())
    }
}
